% Constants
pT = .4;
pS = .2;
pG = .2;

% data array indexes
IDX_USER = 1;
IDX_ACTIVITY = 2;
IDX_RESULT = 3;
IDX_DATE = 4;
IDX_SUBACTIVITY = 5;

COMPUTATION_MODE = 'w'; % l - loose s - strict, 1 - strict once [X] no/go, w - weighted

PROBLEM_FILE_PREFIX = 'sqlknot_';
EXAMPLE_FILE_PREFIX = 'webex_sql_';

% 831; - no examples
groups = [404;535;592;759;760;850];

steps = 10; % number of steps to blend problem with examples

SDATA{size(groups,1),1} = struct([]);
DATA{size(groups,1), steps+1} = struct([]);

time = tic;

% matlabpool open local 4;
%=================
for index=1:size(groups,1),
%==================
	source_data = struct;
	source_data.group_id = groups(index);
	source_data.problem_id = [];
	source_data.problem_name = {};
	source_data.example_id = [];
	source_data.example_name = {};
	source_data.example_no_subs = [];
	source_data.concept_id = [];
	source_data.concept_name = {};
	source_data.user_id = [];
	source_data.user_name = {};
	source_data.problem_log = []; % problem user activity data
	source_data.example_log = []; % example user activity data
	source_data.log_align = []; % alignment of problem-example logs
	source_data.problem_concept = []; % activity-problem incidency
	source_data.problem_concept_dir = []; % activity-problem direction incidency
	source_data.example_concept = []; % example-problem incidency
	source_data.example_concept_dir = []; % example-problem direction incidency

    i_group = groups(index);
    s_group = num2str(i_group);

    [source_data.problem_id, source_data.problem_name] = textread(['data/',PROBLEM_FILE_PREFIX,s_group,'_a.txt'],'%d %s','delimiter', '\t');
    [source_data.concept_id, source_data.concept_name] = textread(['data/',PROBLEM_FILE_PREFIX,s_group,'_c.txt'],'%d %s','delimiter', '\t');
    [source_data.user_id, source_data.user_name, raw1] = textread(['data/',PROBLEM_FILE_PREFIX,s_group,'_u.txt'],'%d %s %d','delimiter', '\t');

    [source_data.example_id, source_data.example_name, source_data.example_no_subs] = textread(['data/',EXAMPLE_FILE_PREFIX,s_group,'_a.txt'],'%d %s %d','delimiter', '\t');
	
	% matrix dimensions
	n_users = size(source_data.user_id, 1);
	n_probl = size(source_data.problem_id, 1);
	n_concept = size(source_data.concept_id, 1);
	n_exampl = size(source_data.example_id, 1);
	

    % user, activity, result, date 
    [a_usr,a_act,a_res,raw4] = textread(['data/',PROBLEM_FILE_PREFIX,s_group,'_d.txt'],'%d %d %d %s','delimiter', '\t');
    [tf, a_usr_true] = ismember(a_usr,source_data.user_id);
    [tf, a_act_true] = ismember(a_act,source_data.problem_id);
    
    source_data.problem_log = [a_usr_true,a_act_true,a_res, datenum(raw4,'yyyy-mm-dd HH:MM:SS.FFF')];
    source_data.problem_log = sortrows(source_data.problem_log,4);

	[a_usr,a_act,a_res,raw4, a_subact] = textread(['data/',EXAMPLE_FILE_PREFIX,s_group,'_d.txt'],'%d %d %f %s %d','delimiter', '\t');
    [u_tf, a_usr_true] = ismember(a_usr, source_data.user_id);
    [tf, a_act_true] = ismember(a_act, source_data.example_id);
    
    source_data.example_log = [a_usr_true,a_act_true,a_res, datenum(raw4,'yyyy-mm-dd HH:MM:SS.FFF'), a_subact];
	source_data.example_log = source_data.example_log(find(u_tf),:);
    source_data.example_log = sortrows(source_data.example_log,4);
	
	% prepare alignment of PROBLEM and EXAMPLE data - #, date, type
	size_p = size(source_data.problem_log,1);
	size_e = size(source_data.example_log,1);
	
	align_Q = [(1:size_p)', source_data.problem_log(:,IDX_DATE), ones(size_p,1)];
	align_W = [(1:size_e)', source_data.example_log(:,IDX_DATE), 2*ones(size_e,1)];
	source_data.log_align = sortrows([align_Q; align_W],2); % alignment

	% reading problem,example-concept indixes
    [a_act,raw1,a_con, raw2, rel_ac_w, rel_actcon_dir] = textread(['data/',PROBLEM_FILE_PREFIX,s_group,'_ac.txt'],'%d %s %d %s %d %d','delimiter', '\t');
    [tf, rel_actcon_act] = ismember(a_act,source_data.problem_id);
    [tf, rel_actcon_con] = ismember(a_con,source_data.concept_id);
	
    [a_act,raw1,a_con, raw2, Wrel_ac_w, Wrel_actcon_dir] = textread(['data/',EXAMPLE_FILE_PREFIX,s_group,'_ac.txt'],'%d %s %d %s %d %d','delimiter', '\t');
    [tf, Wrel_actcon_act] = ismember(a_act,source_data.example_id);
    [tf, Wrel_actcon_con] = ismember(a_con,source_data.concept_id);
	
	size_p = size(source_data.problem_id,1);
	size_e = size(source_data.example_id,1);
	size_c = size(source_data.concept_id,1);
	
    source_data.problem_concept = zeros(size_p, size_c);
    source_data.problem_concept_dir = zeros(size_p, size_c);
    for i=1:size(rel_actcon_act,1),
        source_data.problem_concept( rel_actcon_act(i),rel_actcon_con(i) ) = rel_ac_w(i);
        source_data.problem_concept_dir( rel_actcon_act(i),rel_actcon_con(i) ) =  rel_actcon_dir(i);
    end;
	
    source_data.example_concept = zeros(size_e, size_c);
    source_data.example_concept_dir = zeros(size_e, size_c);
    for i=1:size(Wrel_actcon_act,1),
		% SKIP EXAMPLE CONCEPTS NOT IN PROBLEMS
		if(and(Wrel_actcon_act(i)>0,Wrel_actcon_con(i)>0)),
			source_data.example_concept( Wrel_actcon_act(i), Wrel_actcon_con(i) ) = Wrel_ac_w(i);
			source_data.example_concept_dir( Wrel_actcon_act(i), Wrel_actcon_con(i) ) =  Wrel_actcon_dir(i);
		end;
    end;
	
	SDATA{index} = source_data;
	% === ALL DATA IS READ AT THIS POINT
	
 	parfor i_s=1:(steps+1), %%% FOR ALL STEPS
			t0 = tic;
			
			model_data = struct;
			% user x concept model
			model_data.cum =				zeros( n_users, n_concept );
			model_data.kt =					ones(  n_users, n_concept )*.5; 
			% user x problem model
			model_data.cum_solves =			zeros( n_users, n_probl   );
			model_data.kt_solves =			zeros( n_users, n_probl   ); 
			% learning opportunities all / correct
			model_data.con_lrnop =			zeros( n_users, n_concept ); 
			model_data.con_lrnop_corr =		zeros( n_users, n_concept ); 
			model_data.probl_lrnop =		zeros( n_users, n_probl   ); 
			model_data.probl_lrnop_corr =	zeros( n_users, n_probl   ); 
			model_data.examp_lrnop =		zeros( n_users, n_exampl  ); 
			model_data.examp_explr_subs =	cell(  n_users, n_exampl  ); 
			% logs for predictive validity by concept / problem
			model_data.usr_log =			cell ( n_users, 1		  ); 
			model_data.probl_log =			cell ( n_probl, 1		  ); 
			model_data.usr_con_log =		cell ( n_users, n_concept ); 
			model_data.usr_probl_log =		cell ( n_users, n_probl   ); 
% 			% update volume
% 			model_data.cum_con_updvol =		zeros( n_users, n_concept ); 
% 			model_data.kt_con_updvol =		zeros( n_users, n_concept ); 
% 			model_data.cum_probl_updvol =	zeros( n_users, n_probl   ); 
% 			model_data.kt_probl_updvol =	zeros( n_users, n_probl   ); 
% 			model_data.cum_examp_updvol =	zeros( n_users, n_exampl  ); 
% 			model_data.kt_examp_updvol =	zeros( n_users, n_exampl  ); 

			mode = 0; % 1 - problem, 2 - example
			% compute pL
			for j=1:size(source_data.log_align,1), % for all data 

				mode = source_data.log_align(j,3);
				IDX = source_data.log_align(j,1);

				if(mode==1),
					idx_u = source_data.problem_log(IDX, IDX_USER);
					idx_a = source_data.problem_log(IDX, IDX_ACTIVITY);
					res = source_data.problem_log(IDX, IDX_RESULT);
					idx_c = find( source_data.problem_concept(idx_a,:) )';
					% CUMLATE ONLY
					idx_c_dir = source_data.problem_concept_dir(idx_a,idx_c)';
					idx_c_wght = source_data.problem_concept(idx_a,idx_c)';
				elseif(mode==2),
					idx_u = source_data.example_log(IDX, IDX_USER);
					idx_a = source_data.example_log(IDX, IDX_ACTIVITY);
					idx_suba = source_data.example_log(IDX, IDX_SUBACTIVITY);
					res = abs(source_data.example_log(IDX, IDX_RESULT));
					idx_c = find( source_data.example_concept(idx_a,:) )';
					% CUMLATE ONLY
% disp(['a,c:',num2str(idx_a),' ']);
% disp([num2str(isempty(idx_c))]);
					idx_c_dir = source_data.example_concept_dir(idx_a,idx_c)';
					idx_c_wght = source_data.example_concept(idx_a,idx_c)';
				end;

				out_w_sum = sum(idx_c_wght(find(idx_c_dir==1),:));
				pre_w_sum = sum(idx_c_wght(find(idx_c_dir==0),:));

				if(mode==1), % VALIDITY ONLY FOR PROBLEMS
					% OLD >>> predictive validity: group, user, activity, actual, cumulate, ktrace
					% lookup instead of computation
					res_cum = model_data.cum_solves(idx_u,idx_a);
					res_kt = model_data.kt_solves(idx_u,idx_a);

					% update predictive validity data
					% old tuple [group, user, activity(problem/example), actual, cum_predicted, kt_predicted]
					% new problem tuple [actual, cum_predicted, kt_predicted]
					% new concept tuple [actual, cum_predicted, kt_predicted]
					
					% problem PV log - one tuple per log
					model_data.usr_probl_log{idx_u,idx_a} = [model_data.usr_probl_log{idx_u,idx_a}; res, res_cum, res_kt];
					% concept PV log - multiple tuple per log
					model_data.usr_con_log(idx_u,idx_c) = cellfun(@(x)[x; res, res_cum, res_kt], model_data.usr_con_log(idx_u,idx_c), 'UniformOutput',false);
					% example ; b= cellfun(@(x)[x,1],a,'UniformOutput',false)
					model_data.usr_log{idx_u} = [model_data.usr_log{idx_u}; res, res_cum, res_kt];
					model_data.probl_log{idx_a} = [model_data.probl_log{idx_a}; res, res_cum, res_kt];
					
					
					% OLD group_pv{i_s} = [group_pv{i_s}; i_group, voc_user_id(idx_u), voc_activity_id(idx_a), res, res_cum, res_kt]; *** update PV logs ***

					% if there are no concepts, then no result - continue
					if(isnan(res_cum)),
						continue;
						%res_cum = 0;
					end;
				end;% VALIDITY ONLY FOR PROBLEMS

				exaple_qualify = 0;
				if(mode==2),
					% count explored sub-line
					model_data.examp_explr_subs{idx_u,idx_a} = union(model_data.examp_explr_subs{idx_u,idx_a}, idx_suba);
					if(COMPUTATION_MODE=='l'), % loose
						exaple_qualify = model_data.examp_lrnop(idx_u,idx_a) / source_data.example_no_subs(idx_a) >= .5; 
					elseif(COMPUTATION_MODE=='s'), % strict
						exaple_qualify = size( model_data.examp_explr_subs{idx_u,idx_a},2 )/source_data.example_no_subs(idx_a) >= .5;
					elseif(COMPUTATION_MODE=='1'), % strict, count ONCE
						exaple_qualify = 0;
						if(~ismember(-1,model_data.examp_explr_subs{idx_u,idx_a})),
							exaple_qualify = size( model_data.examp_explr_subs{idx_u,idx_a},2 )/source_data.example_no_subs(idx_a) >= .5;
							if(exaple_qualify), % disable further stuff
								model_data.examp_explr_subs{idx_u,idx_a} = union(model_data.examp_explr_subs{idx_u,idx_a}, -1); % -1 disables
							end;
						end;
					elseif(COMPUTATION_MODE=='w'), % weighted
						exaple_qualify = 1;
					end;
				end;
				
				% USED TO BE AT X1 - check later
				if( and(mode==2, or(~exaple_qualify,isempty(idx_c))) ),
					continue;
				end;
				
				% learning opportunities update
				if(mode==1),
					model_data.probl_lrnop     (idx_u,idx_a) = model_data.probl_lrnop     (idx_u,idx_a) + 1; % inc problem attempt
					model_data.probl_lrnop_corr(idx_u,idx_a) = model_data.probl_lrnop_corr(idx_u,idx_a) + res; % inc problem attempt
				elseif(mode==2),
					model_data.examp_lrnop(idx_u,idx_a) = model_data.examp_lrnop(idx_u,idx_a) + 1; % inc problem attempt
% 					if(exaple_qualify),
% 						ncw5(idx_u,idx_a) = ncw5(idx_u,idx_a) + 1; ***
% 					end;
				end;
                model_data.con_lrnop     (idx_u,idx_c) = model_data.con_lrnop     (idx_u,idx_c) + 1;
                model_data.con_lrnop_corr(idx_u,idx_c) = model_data.con_lrnop_corr(idx_u,idx_c) + res;

				% X1 used to be other code here

                % CUMULATE part
% 				if(mode==1),
					w_sum = ones(size(idx_c)) * pre_w_sum;
% 				elseif(mode==2),
% 					w_sum = ones(size(idx_c)) * 0;
% 				end;
				
				w_sum = w_sum .* (idx_c_dir~=1);
                w_sum = w_sum + (idx_c_dir==1) * out_w_sum;
                multiplier = ones(size(idx_c)) - ( model_data.cum(idx_u,idx_c)'<.5 )*.5;

				cum_update = 0;
                if(mode==1),
                    weight = power(idx_c_wght,.25) ./ power( ( model_data.probl_lrnop_corr(idx_u,idx_a)+1) * w_sum, .25);
					% update volumes
					cum_update = res * weight' .* power( 1-model_data.cum(idx_u,idx_c),2) .* multiplier';
% 					model_data.cum_probl_updvol(idx_u,idx_a) = model_data.cum_probl_updvol(idx_u,idx_a) + mean(cum_update); 
                elseif(mode==2),
                    weight = power(idx_c_wght,.25) ./ power( (model_data.examp_lrnop(idx_u,idx_a)+1)*w_sum, .25);
					cum_update = ((i_s-1)/steps) .* weight' .* power( 1-model_data.cum(idx_u,idx_c),2) .* multiplier';
					
					if(((i_s-1)/steps)>0)
						e=0;
					end;
					
% 					model_data.cum_examp_updvol(idx_u,idx_a) = model_data.cum_examp_updvol(idx_u,idx_a) + mean(cum_update); 
                end;
				% update model
				model_data.cum(idx_u,idx_c) = model_data.cum(idx_u,idx_c) + cum_update;
				% update volumes 
% 				model_data.cum_con_updvol(idx_u,idx_c) = model_data.cum_con_updvol(idx_u,idx_c) + cum_update; 
				
                % KTrace part
                partial_credit = 1;
                if(mode==2),
                    partial_credit = ((i_s-1)/steps);
                    if(COMPUTATION_MODE=='w'), % weight even more
                        partial_credit = partial_credit * size( model_data.examp_explr_subs{idx_u,idx_a},2 )/source_data.example_no_subs(idx_a);
                    end;
                end;

                pLn_1 = model_data.kt(idx_u,idx_c);

                pLn_1_bar_evidence = (res~=0) * pLn_1 * (1 - pS) ./ ( pLn_1 * (1 - pS) + (1 - pLn_1) * pG ) + ...
							(res==0) * pLn_1 * pS  ./ ( pLn_1 * pS + (1 - pLn_1) * (1 - pG) );

                model_data.kt(idx_u,idx_c) = pLn_1_bar_evidence + (1 - pLn_1_bar_evidence) * pT * partial_credit;
				% update volumes
% 				model_data.kt_con_updvol(idx_u,idx_c) = model_data.kt_con_updvol(idx_u,idx_c) + abs((1 - pLn_1_bar_evidence) * pT * partial_credit);
% 				if(mode==1),
% 					model_data.kt_probl_updvol(idx_u,idx_a) = model_data.kt_probl_updvol(idx_u,idx_a) + mean(abs((1 - pLn_1_bar_evidence) * pT * partial_credit));
% 				elseif(mode==2),
% 					model_data.kt_examp_updvol(idx_u,idx_a) = model_data.kt_examp_updvol(idx_u,idx_a) + mean(abs((1 - pLn_1_bar_evidence) * pT * partial_credit));
% 				end;
				
				% recompute probabilities to solve problems right
				if(mode==1),
					idx_p = idx_a;
				elseif(mode==2),
					idx_p = find(sum(source_data.problem_concept(:,idx_c),2)>0);
				end;

				for problem_i = 1:size(idx_p,1),
					idx_c_p = find( source_data.problem_concept(idx_p(problem_i),:) )';
					idx_c_p_dir = source_data.problem_concept_dir(idx_p(problem_i),idx_c_p)';
					idx_c_p_wght = source_data.problem_concept(idx_p(problem_i),idx_c_p)';
					
					model_data.cum_solves(idx_u, idx_p(problem_i)) = prod( ...
						( model_data.cum(idx_u,idx_c_p(find(idx_c_p_dir==1))) .* ...
						  idx_c_p_wght(find(idx_c_p_dir==1))'  ...
						) / ...
						sum(idx_c_p_wght(find(idx_c_p_dir==1))) );

					model_data.kt_solves(idx_u, idx_p(problem_i)) = prod( ...
						( model_data.kt(idx_u,idx_c_p(find(idx_c_p_dir==1))) .* ...
						  idx_c_p_wght(find(idx_c_p_dir==1))'  ...
						) / ...
						sum(idx_c_p_wght(find(idx_c_p_dir==1))) );
				end;
% 					model_data.cum_solves(idx_u, idx_a) = mean( model_data.cum(idx_u,idx_c) );
% 					model_data.kt_solves (idx_u, idx_a) = prod( model_data.kt(idx_u,idx_c)  );
				
			end; % -- for all data 
			
			disp(['group ',s_group,' step ', num2str(i_s), ' t:',num2str(toc(t0))]);
			
			% save step
			DATA{index, i_s} = model_data;
	end; %%% FOR ALL STEPS
end;
% ==================
% matlabpool close;

save('blendmodel_weighted_sql_1.mat','DATA');
save('sql_SDATA.mat','SDATA');
disp(['-- done in ',num2str(toc(time)),' --']);
beep;


% user_stats = [{'group'}, {'user'}, {'u_login'}, {'no_trans'}, {'no_correct'}, {'no_examples'}, {'no_examples_qual'},  ...
% 	{'posrate_act'} ,{'posrate_cum'}, {'posrate_kt'}, {'accuracy_cum'}, {'accuracy_kt'}, {'auc_roc_cum'},...
% 	{'auc_pr_cum'}, {'auc_roc_kt'}, {'auc_pr_kt'}; user_stats];
% 
% conc_stats = [{'group'}, {'concept'}, {'conc_login'}, {'no_trans'}, {'no_correct'}, {'no_examples'}, {'no_examples_qual'},  ...
% 	{'posrate_act'} ,{'posrate_cum'}, {'posrate_kt'}, {'accuracy_cum'}, {'accuracy_kt'}, {'auc_roc_cum'},...
% 	{'auc_pr_cum'}, {'auc_roc_kt'}, {'auc_pr_kt'}; conc_stats];
% 
% disp('--done--');
% save('user_data_wcurve_weighted_sql.mat','user_stats');
% save('conc_data_wcurve_weighted_sql.mat','conc_stats');

